@@ -83,6 +83,8 @@ typedef struct mod_vhost_ldap_config_t {
 
     char *fallback;                     /* Fallback virtual host */
 
+    int wildcard;                       /* Do wildcard search if host not found */
+
 } mod_vhost_ldap_config_t;
 
 typedef struct mod_vhost_ldap_request_t {
@@ -157,6 +159,7 @@ mod_vhost_ldap_create_server_config (apr_pool_t *p, server_rec *s)
     conf->bindpw = NULL;
     conf->deref = always;
     conf->fallback = NULL;
+    conf->wildcard = MVL_ENABLED;
 
     return conf;
 }
@@ -169,11 +172,7 @@ mod_vhost_ldap_merge_server_config(apr_pool_t *p, void *parentv, void *childv)
     mod_vhost_ldap_config_t *conf =
 	(mod_vhost_ldap_config_t *)apr_pcalloc(p, sizeof(mod_vhost_ldap_config_t));
 
-    if (child->enabled == MVL_UNSET) {
-	conf->enabled = parent->enabled;
-    } else {
-	conf->enabled = child->enabled;
-    }
+    conf->enabled = (child->enabled != MVL_UNSET) ? child->enabled : parent->enabled;
 
     if (child->have_ldap_url) {
 	conf->have_ldap_url = child->have_ldap_url;
@@ -207,6 +206,8 @@ mod_vhost_ldap_merge_server_config(apr_pool_t *p, void *parentv, void *childv)
 
     conf->fallback = (child->fallback ? child->fallback : parent->fallback);
 
+    conf->wildcard = (child->wildcard != MVL_UNSET) ? child->wildcard : parent->wildcard;
+
     return conf;
 }
 
@@ -398,6 +399,17 @@ static const char *mod_vhost_ldap_set_fallback(cmd_parms *cmd, void *dummy, cons
     return NULL;
 }
 
+static const char *mod_vhost_ldap_set_wildcard(cmd_parms *cmd, void *dummy, int wildcard)
+{
+    mod_vhost_ldap_config_t *conf =
+	(mod_vhost_ldap_config_t *)ap_get_module_config(cmd->server->module_config,
+							&vhost_ldap_module);
+
+    conf->wildcard = (wildcard) ? MVL_ENABLED : MVL_DISABLED;
+
+    return NULL;
+}
+
 command_rec mod_vhost_ldap_cmds[] = {
     AP_INIT_TAKE1("VhostLDAPURL", mod_vhost_ldap_parse_url, NULL, RSRC_CONF,
                   "URL to define LDAP connection. This should be an RFC 2255 compliant\n"
@@ -428,6 +440,9 @@ command_rec mod_vhost_ldap_cmds[] = {
 		  "is not found in LDAP database. This option can be used to display"
 		  "\"virtual host not found\" type of page."),
 
+    AP_INIT_FLAG("VhostLDAPWildcard", mod_vhost_ldap_set_wildcard, NULL, RSRC_CONF,
+                 "Set to off to disable wildcard search if the requested hostname is not found."),
+
     {NULL}
 };
 
@@ -440,8 +455,10 @@ static int mod_vhost_ldap_translate_name(request_rec *r)
     char filtbuf[FILTER_LENGTH];
     mod_vhost_ldap_config_t *conf =
 	(mod_vhost_ldap_config_t *)ap_get_module_config(r->server->module_config, &vhost_ldap_module);
+#ifndef HAS_PER_REQUEST_DOCUMENT_ROOT
     core_server_config *core =
         (core_server_config *)ap_get_module_config(r->server->module_config, &core_module);
+#endif
     util_ldap_connection_t *ldc = NULL;
     int result = 0;
     const char *dn = NULL;
@@ -452,6 +469,8 @@ static int mod_vhost_ldap_translate_name(request_rec *r)
     int sleep1 = 1;
     int sleep;
     struct berval hostnamebv, shostnamebv;
+    char **document_root_ptr;
+    int ret = DECLINED;
 
     reqc =
 	(mod_vhost_ldap_request_t *)apr_pcalloc(r->pool, sizeof(mod_vhost_ldap_request_t));
@@ -518,17 +537,19 @@ fallback:
     }
 
     if (result == LDAP_NO_SUCH_OBJECT) {
-	if (strcmp(hostname, "*") != 0) {
-	    if (strncmp(hostname, "*.", 2) == 0)
-		hostname += 2;
-	    hostname += strcspn(hostname, ".");
-	    hostname = apr_pstrcat(r->pool, "*", hostname, NULL);
-	    ap_log_rerror(APLOG_MARK, APLOG_NOTICE|APLOG_NOERRNO, 0, r,
-		          "[mod_vhost_ldap.c] translate: "
-			  "virtual host not found, trying wildcard %s",
-			  hostname);
-	    goto fallback;
-	}
+        if (conf->wildcard == MVL_ENABLED) {
+	    if (strcmp(hostname, "*") != 0) {
+	        if (strncmp(hostname, "*.", 2) == 0)
+		    hostname += 2;
+                hostname += strcspn(hostname, ".");
+                hostname = apr_pstrcat(r->pool, "*", hostname, NULL);
+                ap_log_rerror(APLOG_MARK, APLOG_NOTICE|APLOG_NOERRNO, 0, r,
+		              "[mod_vhost_ldap.c] translate: "
+			      "virtual host not found, trying wildcard %s",
+			      hostname);
+	        goto fallback;
+	    }
+        }
 
 null:
 	if (conf->fallback && (is_fallback++ <= 0)) {
@@ -622,6 +643,7 @@ null:
 	  r->filename = cgi;
 	  r->handler = "cgi-script";
 	  apr_table_setn(r->notes, "alias-forced-type", r->handler);
+	  ret = OK;
 	}
     } else if (r->uri[0] == '/') {
         /* we don't set r->filename here, and let other modules do it
@@ -633,6 +655,8 @@ null:
 	return DECLINED;
     }
 
+#ifndef HAS_PER_REQUEST_DOCUMENT_ROOT
+
     if ((r->server = apr_pmemdup(r->pool, r->server, sizeof(*r->server))) == NULL) {
         ap_log_rerror(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, 0, r, 
                       "[mod_vhost_ldap.c] translate: "
@@ -661,9 +685,12 @@ null:
                       "translate failed; Unable to copy r->core structure");
         return HTTP_INTERNAL_SERVER_ERROR;
     }
-    ap_set_module_config(r->server->module_config, &core_module, core);
+    ap_set_module_config(r->server->module_config, &core_module, core);    
 
-    /* Stolen from server/core.c */
+    document_root_ptr = (char **)&core->ap_document_root;
+#else
+    document_root_ptr = (char **)&r->document_root;
+#endif
 
     /* Make it absolute, relative to ServerRoot */
     reqc->docroot = ap_server_root_relative(r->pool, reqc->docroot);
@@ -676,18 +703,18 @@ null:
     }
 
     /* TODO: ap_configtestonly && ap_docrootcheck && */
-    if (apr_filepath_merge((char**)&core->ap_document_root, NULL, reqc->docroot,
+    if (apr_filepath_merge(document_root_ptr, NULL, reqc->docroot,
                            APR_FILEPATH_TRUENAME, r->pool) != APR_SUCCESS
         || !ap_is_directory(r->pool, reqc->docroot)) {
 
         ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, r,
 		      "[mod_vhost_ldap.c] set_document_root: Warning: DocumentRoot [%s] does not exist",
 		      reqc->docroot);
-        core->ap_document_root = reqc->docroot;
+        *document_root_ptr = reqc->docroot;
     }
 
-    /* Hack to allow post-processing by other modules (mod_rewrite, mod_alias) */
-    return DECLINED;
+    /* Return DECLINE to allow post-processing by other modules (mod_rewrite, mod_alias) */
+    return ret;
 }
 
 #ifdef HAVE_UNIX_SUEXEC
